# $Id: Setpwd.pm 32787 2010-10-19 14:34:40Z anton $
# $URL: https://infix.uvt.nl/its-id/trunk/sources/pwdmodifier/keymaster/lib/Setpwd.pm $
package Setpwd;
use strict;
use warnings  FATAL=> 'all';
use Data::Dumper;
use Baseobject;
use Carp;

use Encode qw(encode);

use MIME::Base64;
use Digest::SHA1;
use FileHandle;
use Fcntl qw(:DEFAULT :flock);

use IO::Seekable;
use Lock;

use Rcrypt;

our @ISA=('Baseobject' );
our @EXPORT = qw (verbosity);

use Carp;

my $FAIL="keymaster status: Failure:";

sub new
{
  my $pkg=shift;
  my $self= bless (new Baseobject(@_),$pkg);
  #
  # NOTE NOTE NOTE allow all fields is set to  YES
  #
  $self->{allowallfields}=1;

  $self->needs(
	       [ qw (
		      ipaddress anr ldaphost clearout
		      sn uid nosvnfile ldapservers capath oracle_home  userinfo
		      oracle_pwd_read_table oracle_pwd_write_table allowpwdrepeat sql passwordLeader
		      ) ]
	       );

  $self->allows(
	  [ qw (challenge time oracle_testanr mailalso2 skiporacle dryrun untouchables newPwd
            silentsshclients silentclients sync-passwords reset-password contact language
		     syncOnlyIfNeeded extrabind adminMailAddress) ]
	  );

  $self->allows({'tildeusername' => ''});

  warn "coming here";

  my $cf=shift;
  $self->{cf} = $cf;
  $self->fields(@_);
  $self->fields($self->getparms($cf));
  $self->checkneeds();
  return $self if ($self->error());
  warn "coming here";

  $self->getaccounts();
  return $self if ($self->error());

  warn "coming here";
  $self->{fields}->{allowpwdrepeat}=1 if $self->{stealthmode};

  warn "coming here";
  $self->saveLdapObjects();
  warn "coming here";
  $self->checkAccountInSQL();

  return $self;
}

sub error {
  my $self=shift;

  if (@_ && $_[0]) {
    $self->{error}.= join ('',@_);
    die ("@_\n");
  }
  else  {
    my ($package,$file,$line)=caller;
    return "$package: ",$self->{error};
  }
}

sub extrabind {
    # Verify that the new password can actually be used

	my $self = shift;
	my $port =  636; #fixme!
	my $pwd = $self->{fields}->{newPwd};

	foreach my $host (keys %{$self->{ldapobjects}}) {
		my $ldap;
		$ldap=Net::LDAPS->new(
			$host,
#           FIXME: verify uitgezet voor AD
#           'verify' => 'require',
			'capath' => $self->{fields}->{capath},
			'port'   => $port,
			);

		unless ($ldap) {
			warn "ERROR Could not connect to $host:$port $@";
			next;
		}
		warn "Connection to $host OK";
		foreach my $entry ( @{$self->{ldapobjects}->{$host}->{entries}}){
			my $dn = $entry->dn;
			my $objectclass = $entry->get_value('objectclass');
			if ($objectclass eq 'aliasObject') {
				$self->debug( "skipping aliasObject for: $dn on server $host");
				next;
			}

			my $msg = $ldap->bind( $dn, password=>$pwd );
			if ($msg->code)	{
				warn "    bind with new password failed for: $dn on ldapserver: $host:$port ",$msg->error();
				$self->error('$verificationBindFailed');
			}else {
				warn "    bind OK for dn: '$dn' ";
			}
		}
	}
}

sub saveLdapObjects {
  my $self=shift;

  foreach my $host ( keys %{$self->{ldapaccounts}}) {
    $self->debug ("host $host");
    unless (exists ($self->{ldapaccounts}->{$host}->{protocol})) {
      $self->verbose( "IGNORING $host");
      next ;
    }
    $self->saveldapobject($host);
  }
}


sub fetchCurrentSHAPassword {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	return $sql->getCurrentSHAPassword($anr);
}

sub fetchCurrentRcryptPassword {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	return $sql->getCurrentRcryptPassword($anr);
}

sub fetchCurrentSQLPassword {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $rcryptPwd = $self->fetchCurrentRcryptPassword();
	my $decrypted = Rcrypt::decrypt($rcryptPwd, $self->{rcryptkey});
	warn  "Current SQL Password NOT retreived for $anr" unless ($decrypted);
	return $decrypted;
}

sub synchronisePassword {
	my $self=shift;
	my $ret;

	$self->{fields}->{allowpwdrepeat} = 1;
	$self->{fields}->{skiporacle} = 1 if(($self->{fields}->{passwordLeader} eq 'oracle'));

	unless ( grep /^$self->{fields}->{passwordLeader}$/, qw (oracle ldap rcrypt)){
		die "$FAIL incorrect passwordleader: $self->{fields}->{passwordLeader}";
	}
	my $passwordLeader = $self->{fields}->{passwordLeader};

	$ret->{oracle} = $self->fetchCurrentSQLPassword();
	$ret->{ldap} = $self->fetchCurrentLdapPassword();
	$ret->{rcrypt} = $self->fetchRcryptPassword();

	$self->{fields}->{newPwd} = $ret->{$passwordLeader};

	if($self->{fields}->{syncOnlyIfNeeded}){
		if(exists ($ret->{rcrypt}) and exists ($ret->{oracle}) and exists ($ret->{ldap})){
			if(($ret->{oracle} eq $ret->{ldap}) and ($ret->{oracle} eq $ret->{rcrypt})) {
				warn "No need to sync according to baseldapserver and Oracle";
				return;
			}
		}
	}
	$self->setPassword();
}


sub fetchCurrentLdapPassword {
	# Fetch the current password from the 'ldaphost'
	my $self=shift;
	my $ldaphost = $self->{fields}->{ldaphost};
	#fixme
	$ldaphost =~ s{ldaps?://}{};

	warn "ldaphost: $ldaphost";
	my $ldapobject = $self->{ldapobjects}->{$ldaphost};
	warn Dumper ($ldapobject);

	my $entries =  $ldapobject->{entries};
	my $entry = @$entries[0];

	my $rcryptpwd =	$entry->get_value('rcryptpassword');
	die "NO RCRYPTPWD FOUND for: $self->{fields}->{anr}" unless defined($rcryptpwd);
	die "INVALID RCRYPTPWD found " unless ($rcryptpwd=~s/^\{rcrypt\}//);

	my $decrypted = Rcrypt::decrypt($rcryptpwd,$self->{rcryptkey});

#	die  "Current ldapPassword NOT retreived " unless ($decrypted);
	$self->debug( "Current LDAP password retreived");
	return ($rcryptpwd, $decrypted);
}


sub getparms
{
	my $self=shift;
	my $cf=shift;
	my $h;
	my $cfhash=$cf->hash();

	foreach my $el (keys %$cfhash) {
		$h->{$el} = $cfhash->{$el};
	}

	my $userinfo = $self->{fields}->{userinfo};
	foreach my $el (keys %$userinfo )  {
		$h->{$el} = $userinfo->{$el};
	}
#   fixme:
	$h->{requestfrom}=join ("\n", map {s/\n//; $_;} $cf->value('requestfrom'))
		if $cfhash->{requestfrom};
	$h->{'ldapservers'}= [ ( $cf->value('ldapservers'))];
	$self->error("Config error ",$cf->error()) if ($cf->error());

	#optional untouchables, silent clients etc
	my $untouchables=[ ( $cf->value('untouchables'))];

	$cf->clearerror() if ($cf->error());

	if ($untouchables->[0])
	{
		$h->{'untouchables'} = $untouchables;
		$self->debug( "untouchables: \"@{$h->{untouchables}}\"" ) ;
	}

	my $silentsshclients=[ ( $cf->value('silentsshclients'))];
	$cf->clearerror() if ($cf->error());

	if ($silentsshclients->[0])
	{
		$h->{silentsshclients}= $silentsshclients ;
		$self->debug( "Silentsshclients: \"@{$h->{silentsshclients}}\"" ) ;
	}

	my $silentclients=[ ( $cf->value('silentclients'))];

	if ($silentclients->[0])
	{
		$h->{silentclients}= $silentclients ;
		$self->debug( "Silentclients: \"@{$h->{silentclients}}\"" ) ;
	}

	return $h;
}

sub getaccounts
{

	warn "coming here";

	my $self=shift;
	my $secretfile ="$main::SYSCONFDIR/$self->{fields}->{nosvnfile}";;

	my $secretcf=new Readconfig (
		{ 'configfile'=>$secretfile,
		  'needs' => [qw (ldapaccounts rcryptkey rcryptkey2 oracleaccount) ],
		  'allows' => [ '.*' ],
		  'secret'=>1
		});

	if ($secretcf->error())
	{
		$self->error($secretcf->error()) ;
		return;
	}
	warn "coming here";

	# rcryptkeys
	$self->{rcryptkey}=$secretcf->value('rcryptkey');
	$self->{rcryptkey2}=$secretcf->value('rcryptkey2');

	# privileged ldap accounts
	foreach my $hostspec ( $secretcf->value('ldapaccounts'))
	{
		next unless  ($hostspec);
		my ($host,$account,$pwd)=split(/\s*;\s*/,$hostspec,3);
		$self->{ldapaccounts}->{$host}->{account}=$account;
		$self->{ldapaccounts}->{$host}->{pwd}=$pwd;
	}
	warn "coming here";

	my $oracle;
	($oracle->{host},$oracle->{account},$oracle->{pwd})=split(/\s*;\s*/, $secretcf->value('oracleaccount'));

	die $secretcf->error() if ($secretcf->error());
	warn "coming here";

	$self->{oracleaccount} = $oracle;
	#combine this with the other configfile, represented in fields:

	# restrict possible subkeys, create hash from list
	my $subkeys;
	map {$subkeys->{$_}=1;}  qw (host attribs base testaccount onusernotfound hashtype deletetype
			       uid limit2 onserverdown ldapidentifier additionalfilter bindmethod passwordfield);

	my $servers;
	warn "coming here";

	#
	# use some unique delimitor
	#
	my $delim='{}';
	my $delimexpr=$delim; $delimexpr=~ s/(\W)/\\$1/g;

	my $ldapservers= join ($delim, @{$self->{fields}->{ldapservers}});
#  warn "ldapservers: $ldapservers";
	foreach my $xhost (split /\s*;\s*/ , $ldapservers)
	{
		my $h;
		foreach my $keypair (split /\s*$delimexpr\s*/ , $xhost)
		{
			next unless  ($keypair);
			my ($key,$value)=split(/\s*=\s*/, $keypair,2);
			if (exists( $subkeys->{$key}))
			{
				$h->{$key}=$value;
			}
			else
			{
				die "Error: invalid subkey: \"$key\" of pair: \"$keypair\" in: $xhost\nin file: keymaster.cf";
			}
		}
		$servers->{$h->{host}}=$h;
		delete ($h->{host});
	}
	warn "coming here";

	foreach my $xhost (keys %$servers)
	{
		my $port=389;
		my $protocol;
		my $host=$xhost;
		$host="ldap://$host" unless ($host =~ /^ldaps?/i);
		($protocol,$host)=split( /:\/\//, $host);
		($host,$port)= split ( /:/,$host)   if ($host =~ /:/);

		if (exists ($self->{ldapaccounts}->{$host}))
		{
            # warn "saving: ldapaccounts: $host";
			# merge two hashes
			map { $self->{ldapaccounts}->{$host}->{$_}=$servers->{$xhost}->{$_}}  (keys (%{$servers->{$xhost}}));
			$self->{ldapaccounts}->{$host}->{port}=$port;
			$self->{ldapaccounts}->{$host}->{protocol}=$protocol;
		}
		else {
			warn "NO LDAPACCOUNT SPECIFIED in file \"$secretfile\" for host: $host\n";
			die "configError";
		}
	}
	warn "coming here";

	foreach my $host (keys (%{$self->{ldapaccounts}}))
	{
		if ( !exists($self->{fields}->{userinfo}->{studentaccount})) {
			my $limit2 = $self->{ldapaccounts}->{$host}->{limit2};
			if (defined $limit2 && $limit2 eq 'studentaccounts') {
				delete $self->{ldapaccounts}->{$host};
				warn "skipping $host, because $self->{fields}->{userinfo}->{uid} is no studentaccount";
				next ;
			}
		}

		unless ($self->{ldapaccounts}->{$host}->{protocol}) {
			warn "Warning: host: $host in ldapaccounts.cf is not defined in keymaster.cf";
		}
	}
	warn "coming here";

}

sub genPassword {
	my $self = shift;
	my $size = 15;
	my $string;
	open FH, '/dev/urandom'  or die "Couldn't open /dev/urandum!";
	while ($size) {
		read FH, $_, 1;
		next unless /[A-Za-z0-9]/;
		$string .= $_;
		$size--;
	}
	close FH;
	return $string;
}

sub makePassword {
	my $self = shift;
	my ($capital, $lower, $digit,$s);
	do {
		$s = $self->genPassword();
#		warn "genpassword: $s";
		$capital = $s =~ /([A-Z])/;
		$lower = $s =~ /([a-z])/;
		$digit = $s  =~ /\d/;
	} while (!($capital && $lower && $digit));
	return $s;
}


sub shapePassword {
	my ($self, $server, $cleartext) = @_;
	my $password;
	my $deletetype = $self->{ldapaccounts}->{$server}->{deletetype};

	if ($cleartext eq '') {
		if (defined ($deletetype) && $deletetype eq 'delete') {
			return '' ;
		} else {
			$cleartext = $self->makePassword();
		}
	}

	# allow preference for sha, default is ssha
	my $hashtype = $self->{ldapaccounts}->{$server}->{hashtype};
	if (defined $hashtype){
		if ($hashtype eq 'sha') {
			$password = $self->sha($cleartext);
		} elsif ($hashtype eq 'none') {
			$password = $cleartext;
		} elsif ($hashtype eq 'MickeySoft') {
			$password = $self->activeDirectoryPwd($cleartext);
		} elsif ($hashtype eq 'ssha') {
			$password = $self->ssha($cleartext);
		} else {
			warn "undefined hashtype: '$hashtype' using default: ssha";
			$password = $self->ssha($cleartext);
		}
	} else {
		$password = $self->ssha($cleartext);
	}
	return $password;
}

sub setPassword {
	my $self = shift;
	# bewaar huidige password gegevens
	warn "coming here";
	my ($curPwdRcrypted, $curPwdClearText) = $self->fetchCurrentLdapPassword();

	warn "coming here";

	# fixme: dit kan vast wel met een hash
	if ( exists (($self->{fields}->{untouchables}))) {
		if ( grep ( /$self->{fields}->{anr}/, @{$self->{fields}->{untouchables}})) 	{
			warn  "$FAIL $self->{fields}->{anr} is an UNTOUCHABLE! The password remains untouched :) \n";
			exit 0;
		}
	}

	$self->{fields}->{dryrun}='no' unless (exists ($self->{fields}->{dryrun}));
	$self->debug( "Dryrun state: " .$self->{fields}->{dryrun});
	my $newfields;

	$newfields->{rcryptpassword} = $self->rcrypt($self->{fields}->{newPwd}, $self->{rcryptkey});
	$newfields->{datepasswdchanged} = time;

	#prepare oraclechanges, no commit yet
	unless ($self->{fields}->{dryrun} eq 'yes')	{
		$self->prepare_writeSQL($newfields);
		return if ($self->error);
	}

	my $fail=0;
	my $changed={};

	foreach my $server (keys (%{$self->{ldapobjects}}))	{
		last if $fail;
		my $userpassword = $self->{ldapaccounts}->{$server}->{passwordfield} || 'userpassword';

		my $newpwd = $self->{fields}->{newPwd};
		my $shapedpassword = $self->shapePassword($server, $newpwd);
		$newfields->{$userpassword} = $shapedpassword;

		my $ldap = $self->{ldapobjects}->{$server}->{object};
		my $entries = $self->{ldapobjects}->{$server}->{entries};

		$self->debug(Dumper ($newfields));

		foreach my $entry (@$entries) {
			my $dn= $entry->dn();
			my $objectclass = $entry->get_value('objectclass');
			if ($objectclass eq 'aliasObject') {
				$self->debug( "skipping aliasObject for: $dn on server $server");
				next;
			}

#			$self->verbose( "Changes for $dn on server: \U$server");
			foreach my $attrib (split (/\s*,\s*/, lc($self->{ldapaccounts}->{$server}->{attribs}))) {
				unless(exists ($newfields->{$attrib}))	{
					warn ("\tignoring invalid attrib for host $server: $attrib");
					next;
				}
				if ($newfields->{$attrib}) {
					$self->debug("\tchange: $attrib into $newfields->{$attrib}");
					$entry->replace($attrib => $newfields->{$attrib});
				} else {
					$self->debug("\tdelete: $attrib");
					$entry->delete($attrib);
				}
			}
			unless ($self->{fields}->{dryrun} eq 'yes') {
				my $msg = $entry->update($ldap);
				if ($msg->code) {
					warn "Ldap update  ERROR: $server, $dn, ",$msg->error();
					warn Dumper ($entry);
					$fail=1;
				}
				else {
					$self->verbose( "Password changed for $dn on server: \U$server");
					$changed->{$server} = $entries;
				}
			}
			else {
				warn "** NOT CHANGING PASSWORD ** for $dn on server \U$server, because of dryrun";
			}
		} # endof foreach server
	}
	unless ($self->{fields}->{dryrun} eq 'yes')
	{
		if (exists ($self->{fields}->{clearout}))
		{
			$fail=1 unless ($self->clearout($newfields))  ;
		}
	}

#   COMMENT NEXT LINE!
#	$fail = 1;

	if ($fail){
		foreach my $server (keys (%$changed)) {
			warn "RESTORING server: \u$server";
			my $userpassword = $self->{ldapaccounts}->{$server}->{passwordfield} || 'userpassword';
			my $ldap=$self->{ldapobjects}->{$server}->{object};
			my $entryList = $self->{ldapobjects}->{$server}->{entries};

			foreach my $entry (@$entryList){
				my $dn = $entry->dn();
				foreach my $attrib (split (/\s*,\s*/, lc($self->{ldapaccounts}->{$server}->{attribs}))) {
					my $value;
					if($attrib eq $userpassword){
						$value = $self->shapePassword($server, $curPwdClearText);

					} else {
						$value = $entry->get_value($attrib);
						unless (defined ($value)){
							warn "Cannot restore attrib: $attrib for $self->{fields}->{anr}";
						}
					}
					$entry->replace($attrib, $value);
				}

				my $msg = $entry->update($ldap);
				if ($msg->code) {
					warn("Ldap RESTORE error! $server, $dn ", $msg->error());
				}
				else {
					warn "Ldap RESTORE succeeded for: $server, $dn ";
				}
			}
		}

		$self->finalize_write2oracleSQL(0);
		die  "password modification error, quiting";
	}
	else {
		$self->finalize_write2oracleSQL(1);
	}
}


sub saveldapobject
{
	my $self=shift;
	my $host=shift;
#   warn $self->inspect($self->{ldapaccounts});
	$self->debug("saveldapobject host: $host");
	my $port = $self->{ldapaccounts}->{$host}->{port};

	$self->debug ( "port: $port, host:$host");
	my $account = $self->{ldapaccounts}->{$host}->{account};
	my $base    = $self->{ldapaccounts}->{$host}->{base};
	my $pwd     = $self->{ldapaccounts}->{$host}->{pwd};
	my $protocol     = $self->{ldapaccounts}->{$host}->{protocol};

	my $ldap;
	warn "connnecting to $host";
	$ldap=Net::LDAPS->new(
		$host,
#       FIXME: verify uitgezet voor AD
#		'verify' => 'require',
		'capath' => $self->{fields}->{capath},
		'port'   => $port,
		);

	unless ($ldap)
	{
		warn "Could not connect to $protocol://$host:$port $@";
		if ( exists ($self->{ldapaccounts}->{$host}->{onserverdown}) and
			 $self->{ldapaccounts}->{$host}->{onserverdown} eq lc ('skip')
			)
		{
			warn "WARNING: Skipping host: $host, because of 'onserverdown is skip'";

			delete ($self->{ldapaccounts}->{$host});
			delete ($self->{ldapobjects}->{$host});

			return;
		}
		else
		{
			print "$FAIL Could not connect to ldapserver: $protocol://$host";
			exit 1;
		}
	}
	warn "coming here";

	my $bindmethod = $self->{ldapaccounts}->{$host}->{bindmethod};
	my ($dn, $msg, $entry);
	if (defined ($bindmethod && $bindmethod eq 'dn')){
		$dn = $account;

	} else {
		my $accountbase=$base;
		if ($account=~ /,/)
		{
			($account,$accountbase) = split ( /\s*,\s*/ , $account,2);
		}

		# search for the special account
		$msg = $ldap->search(base=>$accountbase, filter=>$account);
		if ($msg->code)	{
			warn ("Error using filter: $account in ldapserver: $protocol://$host, accountbase: $accountbase, ",$msg->error);
			exit 1;
		}

		$entry=($msg->entries())[0];
		unless (defined ($entry)) {
			warn ("User \"$account\" unknown on ldapserver: $protocol://$host, accountbase: $accountbase");
			exit 1;
		}
		$dn=$entry->dn();
	}

	warn "coming here";

	$msg=$ldap->bind( $dn, password=>$pwd );
	if ($msg->code)	{
		die "error binding $dn on ldapserver: $protocol://$host:$port ",$msg->error();
	}

	$self->debug( "successful bind to  ldap server: $host, binder:$dn, port: $port, base:$base");
	$self->{ldapobjects}->{$host}->{object}=$ldap;
	my $filter="employeenumber=$self->{fields}->{anr}";

	# FIXME: gedrag controleren
	if (exists(($self->{ldapaccounts}->{$host}->{limit2}))){
		return if
			($self->{ldapaccounts}->{$host}->{limit2} eq 'students') &&
			(! $self->{fields}->{userinfo}->{studentaccount} );
	}

	my $uid = $self->{fields}->{userinfo}->{uid};
	if ( exists $self->{ldapaccounts}->{$host}->{ldapidentifier}) {
		my $saved = $filter;
		my $identifier = $self->{ldapaccounts}->{$host}->{ldapidentifier};
		$filter = "$identifier=$uid";
		$self->verbose("ldapidentifier CHANGING FILTER from '$saved' with '$filter' for host: $host");
	}

	if ( exists ($self->{ldapaccounts}->{$host}->{testaccount})){
		warn ( "Overruling account: $filter with $self->{ldapaccounts}->{$host}->{testaccount} for host: $host");
		$filter = $self->{ldapaccounts}->{$host}->{testaccount};
	}

	$msg = $ldap->search(
		base=>$base,
		filter=>$filter
		);
	if ($msg->code)
	{
		warn "error using filter: $filter in ldapserver: $host", $msg->error();
		$entry = undef;
	} else {
		$entry=($msg->entries())[0];
	}

	unless (defined ($entry))
	{
#       warn $self->inspect($self->{ldapaccounts}->{$host});
		if ( exists ($self->{ldapaccounts}->{$host}->{onusernotfound}) )
		{
			my $skip=0;
			if ($self->{ldapaccounts}->{$host}->{onusernotfound} eq 'skip')
			{
				$self->verbose( "SKIPPING server: $host, since user: \"$filter\" was not found");
				$skip=1;
			}
			elsif ($self->{ldapaccounts}->{$host}->{onusernotfound} eq 'skipunlessstudent')
			{
				unless (exists($self->{fields}->{userinfo}->{studentaccount})) {
					$skip=1;
					$self->verbose( "SKIPPING server: \U$host, since user: \"$filter\" was not found plus user is not student");
					$self->debug( $self->inspect($self->{fields}->{userinfo}));
				}
			}

			if ($skip)
			{
				delete ($self->{ldapaccounts}->{$host});
				delete ($self->{ldapobjects}->{$host});
				return;
			}
		}
		else
		{
			warn ("filter: \"$filter\" unknown on ldapserver: $host ");
			print ("$FAIL ldap_unknown_user, $filter\n");
			exit 0;
		}
	}
	else
	{
		$self->verbose( "Found user: $filter,  on: $host");

		# save the current attribs for restoring purposes for this user
		$dn=$entry->dn();
		$self->{ldapobjects}->{$host}->{dn}=$dn;
		$self->{ldapobjects}->{$host}->{entries}=[$msg->entries()];
	}

}

# fixme dit is een generieke methode, niet specifiek voor Setpwd
sub checkAccountInSQL {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	my $oUid = $sql->getUsername($anr);

	unless (defined ($oUid)){
		$self->error ("No UID found for anr: '$anr'");
	} else {
		$self->{fields}->{uid} = $oUid;
	}
	return $oUid;
}

sub update_passwordSQL {
  my ($self, $anr, $newfields)=@_;
  my $newPwd = $self->{fields}->{newPwd};

  my $pw_rcrypt_after = $newfields->{rcryptpassword};
  my $pw_sha_after = $self->ssha($newPwd);
  chomp ($pw_sha_after);

  my $fullldaphost = $self->{fields}->{ldaphost};
  my ($ldaphost) = $fullldaphost=~/^[^\/]+\/\/([^:]+)/;

  my $pw_rcrypt_before = $self->fetchCurrentRcryptPassword() || $self->rcrypt( '',$self->{rcryptkey});
  my $pw_sha_before = $self->fetchCurrentSHAPassword() || $self->sha('');

  my $changed_by = $self->{fields}->{userinfo}->{changed_by};
  my $username = $self->{fields}->{uid};
  my $epoch_time = $newfields->{datepasswdchanged};

  my $sql = $self->{fields}->{sql};

  $sql->setPwd( $anr, $username, $pw_sha_before, $pw_rcrypt_before,
				  $pw_sha_after, $pw_rcrypt_after, $changed_by, $epoch_time);
}

sub prepare_writeSQL {
	my $self = shift;
	my $newfields = shift;
	if ($self->{fields}->{skiporacle})  {
		warn "SKIPPING ORACLE on request";
		$self->{skiporacle} = 1;
		return;
	}

	my $anr=$self->{fields}->{anr};
	if (exists ($self->{fields}->{oracle_testanr})){
		$anr = $self->{fields}->{oracle_testanr};
		warn "Overwriting ORACLE anr with '$anr'";
	}

	unless ($self->{fields}->{allowpwdrepeat}) {
		if ($self->usedPasswordBefore()) {
			$self->error("usedpasswordbefore\n");
			return;
		}
	}

    $self->update_passwordSQL($anr, $newfields);
}

sub finalize_write2oracleSQL {
  #state ==1 commit, state==0 rollback
  my ($self,  $state)=@_;

  if ($self->{fields}->{dryrun} eq 'yes') {
	  warn "** NOT CHANGING PASSWORD ** in oracle, because of dryrun";
  }

  my $sql = $self->{fields}->{sql};
  my $anr = $self->{fields}->{anr};
  my $status=0;

  unless ($self->{skiporacle}) {
	  if($state) {
		  $sql->commit;
	  } else {
		  $sql->rollback;
	  }
  } else {
	  warn "Not commiting because of 'skiporacle'";
  }
}

sub usedPasswordBefore {
	my ($self) = @_;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};

	# N.B. previousPasswords returns an array of refs to arrays
	my $a = $sql->previousPasswords($anr);
	foreach my $encr (@$a) {
		return 1
			if $self->{fields}->{newPwd} eq Rcrypt::decrypt($encr,$self->{rcryptkey});
	}
	return 0;
}


####################################################################

sub sha
{
	my ($self, $cleartext) = @_;

	my $ctx = Digest::SHA1->new;
	$ctx->add($cleartext);
	return '{SHA}' . encode_base64($ctx->digest);
}

sub ssha{
	my ($self, $cleartext) = @_;
	my $ctx = Digest::SHA1->new;
	$ctx->add($cleartext);
	my $salt = $self->genrandom(8);
	$salt = encode_base64($salt); chop($salt); chop($salt);
	$ctx->add($salt);
	return '{SSHA}' . encode_base64($ctx->digest . $salt);
}

sub activeDirectoryPwd {
	my ($self, $cleartext) = @_;

	return encode('UTF-16le', '"'.$cleartext.'"' );
}

sub genrandom
{
  my $self=shift;
  my $size=shift || 256;
  my $random;
  if (open FH, '/dev/urandom' )
  {
    read FH, $random, $size;
    close FH;
    return $random;
  }

  #
  # shitty way to generate random string,
  # but we have no /dev/urandom device
  #

  my $max=$size;

  my $min= 0;
  my $maxrange = 255;

  for (my $i=0;$i<$max;$i++)
  {
    $random.= int (rand ($maxrange)) + $min;
  }
  $random;
}

###################################################################
sub rcrypt
{
  my ($self,$cleartext,$key)=@_;
#  warn $cleartext;
#  warn length ($self->{rcryptkey});
  return "{rcrypt}".Rcrypt::encrypt($cleartext,$key);
}
####################################################################

sub clearout
{
  # returns 0 on failure

  my $self=shift;
  my $newfields=shift;

  my $filehandle= new FileHandle();
  my $lock= new Lock (
		      {
			filename=>">>$self->{fields}->{clearout}",
			lockonly=>1,
			lockpath=>$main::LOCALSTATEDIR,
			locktype=>LOCK_EX,
		      }
		      );

  unless ($lock->open())
  {
    warn $self->error($lock->error());
    return 0;
  }

  unless ( -r "$main::LOCALSTATEDIR/$self->{fields}->{clearout}" )
  {
    # if file does not exist in the very beginning only
    unless (sysopen ($filehandle,"$main::LOCALSTATEDIR/$self->{fields}->{clearout}",O_WRONLY|O_CREAT|O_TRUNC,0600))
    {
      $self->stderr("Could not open file for writing: $self->{fields}->{clearout}, $!");
      warn "sysopen failed!, $!";
      return 0;
    }
    close ($filehandle);
  }
  unless (open ($filehandle,"$main::LOCALSTATEDIR/$self->{fields}->{clearout}"))
  {
    $self->stderror("Could not open file for reading: $self->{fields}->{clearout}, $!");
    return 0;
  }

  my ($h,$key,$value);
  #convert the entire file into a hash
  map { chomp; ($key,$value)= split ('\|',$_,2); $h->{$key}=$value  }  (<$filehandle>);

  close ($filehandle);

  # write to the hash while the file is still closed
  $h->{$self->{fields}->{anr}}= "$newfields->{datepasswdchanged}|". $self->rcrypt($self->{fields}->{newPwd},$self->{rcryptkey2});

  unless (sysopen ($filehandle,"$main::LOCALSTATEDIR/$self->{fields}->{clearout}",O_WRONLY|O_CREAT|O_TRUNC,0600))
  {
    $self->stderr("Could not open file for writing: $self->{fields}->{clearout}, $!");
    warn "sysopen failed!, $!";
    return 0;
  }
  else
  {
    #insert the current element into the hash

    #write the entire sorted hash back to the file
    map {print $filehandle "$_\|$h->{$_}\n"} sort keys %$h;
    close ($filehandle);
  }
  $lock->close();
  return 1;
}

sub fetchRcryptPassword {
  # returns 0 on failure
  my $self=shift;
  my $anr=$self->{fields}->{anr};

  my $filehandle= new FileHandle();
  my $lock= new Lock (
		      {
				  filename=>"$self->{fields}->{clearout}",
				  lockonly=>1,
				  lockpath=>$main::LOCALSTATEDIR,
				  locktype=>LOCK_EX,
		      }
		      );

  unless ($lock->open())
  {
    warn $self->stderror($lock->error());
    return 0;
  }

  unless (open ($filehandle,"$main::LOCALSTATEDIR/$self->{fields}->{clearout}"))
  {
    $self->stderror("Could not open file for reading: $self->{fields}->{clearout}, $!");
    return 0;
  }

  my ($h,$key,$value);
  #convert the entire file into a hash
  map { chomp; ($key,$value)= split ('\|',$_,2); $h->{$key}=$value  }  (<$filehandle>);
  close ($filehandle);
  $lock->close();
  if (defined ($h->{$anr}))
  {
    my $res=$h->{$anr};
    $res =~ s/^\d+\|\{rcrypt\}//;
    return Rcrypt::decrypt($res,$self->{rcryptkey2} );
  }
  return 0;
}

42;

__END__
